package com.collegeboard.gridworld.actor;

/**
 *  An <code>Actor</code> is an entity in a world that can
 *  act.
 *
 *  <p>
 *  copyright&copy; 2005 Cay S. Horstmann (http://horstmann.com)
 *
 *  @author Cay Horstmann
 **/

import java.awt.Color;
import java.util.ArrayList;

import com.collegeboard.gridworld.grid.Grid;
import com.collegeboard.gridworld.grid.Location;


public abstract class Actor
{
   private Grid<Actor> grid;
   private Location location;
   private int direction;
   private Color color;
   
   /**
    * Constructs a black actor that is facing north.
    */
   public Actor()
   {
      this.direction = Location.NORTH;
      this.color  = Color.BLACK;
   }

   /**
    * Gets the location of this actor
    * <br>Precondition: This actor is in a grid
    * @return this actor's location
    */
   public Location getLocation()
   {
      return location;
   }
   
   /**
    * Gets the valid locations of the neighbors in all directions
    * (north, northeast, east, southeast, south, southwest, west,
    * and northwest)
    * <br>Precondition: This actor is in a grid
    * @return an array list of the valid locations of loc in all
    * eight directions
    */
   public ArrayList<Location> getValidNeighborLocations()
   {
      ArrayList<Location> nbrs = new ArrayList<Location>();

      int d = Location.NORTH;
      for (int i = 0; i < Location.FULL_CIRCLE / Location.HALF_RIGHT; i++)
      {
         Location neighbor = location.getNeighborLocation(d);
         if (grid.isValid(neighbor))
            nbrs.add(neighbor);
         d = d + Location.HALF_RIGHT;
      }
      return nbrs;
   }
   
   /**
    * Gets the grid in which this actor is located
    * @return the grid, or null if this actor is not contained in a grid
    */
   public Grid<Actor> getGrid()
   {
      return grid;
   }
   
   /**
    * Puts this actor into a grid. If there is another actor at the given location, it is removed.
    * <br>Precondition: This actor is not in a grid
    * @param gr the grid into which this actor should be placed
    * @param loc the location into which the actor should be placed (must be valid)
    */
   public void putSelfInGrid(Grid<Actor> gr, Location loc)
   {
      assert (grid == null);
      
      Actor actor = gr.get(loc);
      if (actor != null) actor.removeSelfFromGrid();
      gr.put(loc, this);
      grid = gr;
      location = loc;
   }
   
   /**
    * Removes an actor from a grid. 
    * <br>Precondition: This actor has been put in a grid.
    */
   public void removeSelfFromGrid()
   {
      assert (grid != null);
      assert (grid.get(location) == this);
      
      grid.remove(location);
      grid = null;
      location = null;
   }
   
   /**
    * Moves this actor to a new location. If there is another actor at the given location, it is removed.
    * <br>Precondition: This actor has been put in a grid
    * @param newLocation the new location (must be valid)
    */
   public void moveTo(Location newLocation)
   {
      assert (grid != null);
      assert (grid.get(location) == this);
      
      grid.remove(location);
      Actor other = grid.get(newLocation);
      if (other != null) other.removeSelfFromGrid();
      location = newLocation;
      grid.put(location, this);
   }
   
   /**
    * Override this method with the action that your actor carries out.actor needs to
    */
   public abstract void act();
   
   /**
    * Turns the direction of this actor.
    * @param angle the angle (clockwise) by which to turn
    */
   public void turn(int angle)
   {
      direction += angle;
      // normalize to { 0, 1, ..., 359 }
      while (direction < 0)
         direction += Location.FULL_CIRCLE;
      direction = direction % Location.FULL_CIRCLE;
   }

   /**
    * Gets the current direction of this actor.
    * @return the direction in { 0, 1, ..., 359 }
    */
   public int getDirection()
   {
      return direction;
   }
   
   /**
    * Gets the color of this actor.
    * @return the color
    */
   public Color getColor()
   {
      return color;
   }
   
   /**
    * Sets the color of this actor.
    * @param newColor the new color
    */
   public void setColor(Color newColor)
   {
      color = newColor;
   }
   
   public String toString()
   {
      return getClass().getName() + "[direction=" + direction
            + ",color=" + color + "]";
   }
}