package com.collegeboard.gridworld.grid;

/**
 *  The <code>BoundedEnv</code> class is
 *  copyright&copy; 2002 College Entrance Examination Board
 *  (www.collegeboard.com).
 *  copyright&copy; 2005 Cay S. Horstmann (www.horstmann.com)
 *
 *  @author Alyce Brady
 *  @author APCS Development Committee
 *  @author Cay S. Horstmann
 **/

import java.util.ArrayList;

public class BoundedGrid<E> implements Grid<E>
{
   private Object[][] occupantArray; // the array storing the grid elements
   
   /**
    * Constructs an empty BoundedGrid object with the given dimensions.
    * (Precondition: <code>rows > 0</code> and <code>cols > 0</code>.)
    * 
    * @param rows
    *           number of rows in BoundedGrid
    * @param cols
    *           number of columns in BoundedGrid
    */
   public BoundedGrid(int rows, int cols)
   {
      if (rows <= 0) throw new IllegalArgumentException("rows <= 0");
      if (cols <= 0) throw new IllegalArgumentException("cols <= 0");
      occupantArray = new Object[rows][cols];
   }

   public int numRows()
   {
      return occupantArray.length;
   }

   public int numCols()
   {
      // Note: according to the constructor precondition, numRows() > 0, so
      // theGrid[0] is non-null.
      return occupantArray[0].length;
   }

   public boolean isValid(Location loc)
   {
      return 0 <= loc.row() && loc.row() < numRows()
            && 0 <= loc.col() && loc.col() < numCols();
   }

   
   public ArrayList<Location> getOccupiedLocations()
   {
      ArrayList<Location> theLocations = new ArrayList<Location>();

      // Look at all grid locations.
      for (int r = 0; r < numRows(); r++)
      {
         for (int c = 0; c < numCols(); c++)
         {
            // If there's an object at this location, put it in the array.
            Location loc = new Location(r, c);
            if (get(loc) != null)
            {
               theLocations.add(loc);
            }
         }
      }

      return theLocations;
   }

   public E get(Location loc)
   {
      if (!isValid(loc))
         throw new IllegalArgumentException("Location " + loc + " is not valid");
      return (E) occupantArray[loc.row()][loc.col()]; // unavoidable warning
   }

   public E put(Location loc, E obj)
   {
      // Check precondition. Location should be valid.
      if (!isValid(loc))
         throw new IllegalArgumentException("Location " + loc + " is not valid");
      if (obj == null)
         throw new NullPointerException();

      // Add object to the grid.
      E oldOccupant = get(loc);
      occupantArray[loc.row()][loc.col()] = obj;
      return oldOccupant;
   }

   public E remove(Location loc)
   {
      if (!isValid(loc))
         throw new IllegalArgumentException("Location " + loc + " is not valid");
      // Remove the object from the grid.
      E r = get(loc);
      occupantArray[loc.row()][loc.col()] = null;
      return r;
   }

   /**
    * Creates a string representing all the objects in this grid
    * (not necessarily in any particular order).
    * 
    * @return a string indicating all the objects in this environment
    */
   public String toString()
   {
      ArrayList<Location> locations = getOccupiedLocations();
      String s = "{";
      for (int index = 0; index < locations.size(); index++)
      {
         if (index > 0)
            s += ", ";
         Location loc = locations.get(index);
         s += loc + "=" + get(loc);
      }
      return s + "}";
   }
}
