package com.collegeboard.gridworld.grid;
// AP(r) Computer Science Marine Biology Simulation:
// The Location class is copyright(c) 2002 College Entrance
// Examination Board (www.collegeboard.com).
//
// This class is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation.
//
// This class is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

/**
 *  AP&reg; Computer Science Marine Biology Simulation:<br>
 *  A <code>Location</code> object represents the row and column of a
 *  location in a two-dimensional grid.
 *
 *  <p>
 *  The <code>Location</code> class is
 *  copyright&copy; 2002 College Entrance Examination Board
 *  (www.collegeboard.com).
 *
 *  @author Alyce Brady
 *  @author Chris Nevison
 *  @author APCS Development Committee
 *  @version 1 July 2002
 **/

public class Location implements Comparable
{
    // Instance Variables: Encapsulated data for each Location object
    private int myRow;            // row location in grid
    private int myCol;            // column location in grid
   
    public static final int FULL_CIRCLE = 360;
    public static final int HALF_CIRCLE = 180;
    public static final int HALF_LEFT = -45;
    public static final int LEFT = -90;
    public static final int HALF_RIGHT = 45;
    public static final int RIGHT = 90;

    public static final int NORTH = 0;
    public static final int NORTHEAST = 45;   
    public static final int EAST = 90;   
    public static final int SOUTHEAST = 135;
    public static final int SOUTH = 180;   
    public static final int SOUTHWEST = 225;   
    public static final int WEST = 270;   
    public static final int NORTHWEST = 315;  
   
    /** Constructs a <code>Location</code> object.
     *  @param row    location's row
     *  @param col    location's column
     **/
    public Location(int row, int col)
    {
        myRow = row;
        myCol = col;
    }

  // accessor methods

    /** Returns the row coordinate of this location.
     *  @return        row of this location
     **/
    public int row()
    {
        return myRow;
    }

    /** Returns the column coordinate of this location.
     *  @return        column of this location
     **/
    public int col()
    {
        return myCol;
    }

    /** Indicates whether some other <code>Location</code> object is
     *  "equal to" this one.
     *  @param other    the other location to test
     *  @return     <code>true</code> if <code>other</code> is at the
     *              same row and column as the current location;
     *              <code>false</code> otherwise
     **/
    public boolean equals(Object other)
    {
        if ( ! (other instanceof Location) )
          return false;

        Location otherLoc = (Location) other;
        return row() == otherLoc.row() && col() == otherLoc.col();
    }

    /** Generates a hash code for this location
     *  (will not be tested on the Advanced Placement exam).
     *  @return     a hash code for a <code>Location</code> object
     **/
    public int hashCode()
    {
       return row() * 3737 + col();
    }

    /** Compares this location to <code>other</code> for ordering.
     *  Returns a negative integer, zero, or a positive integer as this
     *  location is less than, equal to, or greater than <code>other</code>.
     *  Locations are ordered in row-major order.
     *  (Precondition: <code>other</code> is a <code>Location</code> object.)
     *  @param other    the other location to test
     *  @return     a negative integer if this location is less than
     *              <code>other</code>, zero if the two locations are equal,
     *              or a positive integer if this location is greater than
     *              <code>other</code>
     **/
    public int compareTo(Object other)
    {
        Location otherLoc = (Location) other;
        if ( equals(other) )
            return 0;
        if ( row() == otherLoc.row() )
            return col() - otherLoc.col();
        return row() - otherLoc.row();
    }

    /** Represents this location as a string.
     *  @return        a string indicating the row and column of the
     *                 location in (row, col) format
     **/
    public String toString()
    {
        return "(" + row() + ", " + col() + ")";
    }

    /**
     * Gets the neighbor location in the given direction 
     * @param direction an angle (0 degrees = Location.NORTH,
     * 45 degrees = Location.NORTHWEST, etc.)
     * @return the neighbor location in the direction that is obtained by
     * rounding angle to the nearest multiple of 45 degrees
     */
    public Location getNeighborLocation(int direction)
    {
       // reduce mod 360 and round to nearest multiple of 45
       int adjustedDirection = direction;
       while (adjustedDirection < 0)
          adjustedDirection += FULL_CIRCLE;
       adjustedDirection += HALF_RIGHT / 2;
       adjustedDirection = (adjustedDirection / HALF_RIGHT) * HALF_RIGHT;
       int dc = 0;
       int dr = 0;
       if (adjustedDirection == EAST)
          dc = 1;
       else if (adjustedDirection == SOUTHEAST)
       {
          dc = 1;
          dr = 1;
       }
       else if (adjustedDirection == SOUTH)
          dr = 1;
       else if (adjustedDirection == SOUTHWEST)
       {
          dc = -1;
          dr = 1;
       }
       else if (adjustedDirection == WEST)
          dc = -1;
       else if (adjustedDirection == NORTHWEST)
       {
          dc = -1;
          dr = -1;
       }
       else if (adjustedDirection == NORTH)
          dr = -1;
       else if (adjustedDirection == NORTHEAST)
       {
          dc = 1;
          dr = -1;
       }
       return new Location(row() + dr, col() + dc);
    }
    
    /**
     * Returns the direction from this location toward another location
     * @param target another location
     * @return the closest compass direction from this location toward target
     */
    public int directionToward(Location target)
    {
       int dx = target.col() - col();
       int dy = target.row() - row();
       // y axis points opposite to mathematical orientation 
       int angle = (int) Math.toDegrees(Math.atan2(-dy, dx));
       
       // mathematical angle is counterclockwise from x-axis, 
       // compass angle is clockwise from y-axis
       int compassAngle = RIGHT - angle; 
       // prepare for truncating division by 45 degrees
       compassAngle += HALF_RIGHT / 2;        
       // wrap negative angles    
       if (compassAngle < 0) compassAngle += FULL_CIRCLE;
       // round to nearest multiple of 45
       return (compassAngle / HALF_RIGHT) * HALF_RIGHT;       
    }
}
