package com.collegeboard.gridworld.gui;



import java.awt.Rectangle;
import java.beans.*;
import java.lang.reflect.*;
import java.awt.Component;
import java.awt.Graphics2D;
import java.awt.BasicStroke;

/**
 *  AP&reg; Computer Science Marine Biology Simulation:<br>
 *  This class provides common implementation code for drawing fish
 *  using the <code>LocatableDisplay</code> interface. It will translate, 
 *  scale, and rotate the graphics system as needed and then invoke its
 *  abstract <code>draw</code> method. Subclasses of this abstract class 
 *  define <code>draw</code> to just display a fish in a fixed size and
 *  orientation.
 *
 *  <p>
 *  copyright&copy; 2002 College Entrance Examination Board (www.collegeboard.com).
 *  copyright&copy; 2005 Cay S. Horstmann (http://horstmann.com)
 *
 *  @author Julie Zelenski
 *  @author Cay Horstmann
 **/

public abstract class AbstractDisplay implements Display
{    
    
    /** Draw the given Actor.
     *  Subclasses should implement this method to draw the Actor facing North 
     *  in a cell of size (1,1) centered around (0,0) on the drawing surface.
     *  (All scaling/rotating has been done beforehand).
     *  @param   fish       fish we want to draw
     *  @param   comp       component on which to draw
     *  @param   g2         drawing surface
     **/
    abstract public void draw(Object obj, Component comp, Graphics2D g2);
    
    
    /** Draw the given object.
     *  Scales and rotates the coordinate appropriately then invokes
     *  the simple draw method above that is only responsible for
     *  drawing a unit-length Actor facing North.
     *  (Precondition: <code>obj</code> is a Actor (or a Actor subclass).)
     *  @param   obj        object we want to draw
     *  @param   comp       component on which to draw
     *  @param   g2         drawing surface
     *  @param   rect       rectangle in which to draw 
     **/
    public void draw(Object obj, Component comp, Graphics2D g2, Rectangle rect)
    {
        float scaleFactor = Math.min(rect.width, rect.height);
        
        // Translate to center of fish
        g2.translate(rect.x + rect.width/2, rect.y + rect.height/2);
        
        // Rotate drawing surface before drawing to capture fish's
        // orientation (direction).
        if (obj != null)
        {
		     Integer direction = (Integer) getProperty(obj, "direction");
           int rotationInDegrees = direction == null ? 0 : direction.intValue();
           g2.rotate(Math.toRadians(rotationInDegrees));
           if (rotationInDegrees > 180) g2.scale(-1, 1); // flip side-to-side
        }
        // Scale to size of rectangle, adjust stroke back to 1-pixel wide
        g2.scale(scaleFactor, scaleFactor);
        g2.setStroke(new BasicStroke(1.0f/scaleFactor));
    
        draw(obj, comp, g2);
    }
	
	public static Object getProperty(Object obj, String propertyName)
	{
      if (obj == null) return null;
		try
		{
			BeanInfo info = Introspector.getBeanInfo(obj.getClass());
			PropertyDescriptor[] descriptors = info.getPropertyDescriptors();
			for (int i = 0; i < descriptors.length; i++)
			{
			   if (descriptors[i].getName().equals(propertyName))
			   {
				   Method getter = descriptors[i].getReadMethod();
				   if (getter == null) return null;
				   return getter.invoke(obj);
			   }			   
			}
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
		}
		return null;
	}
 }
