package com.collegeboard.gridworld.gui;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;
import java.util.HashMap;

import javax.swing.Icon;

/**
 * AP&reg; Computer Science Marine Biology Simulation:<br>
 * <code>DisplayMap</code> is a collection that maps environment object
 * classes to objects that know how to display them.
 * 
 * <p>
 * copyright&copy; 2002 College Entrance Examination Board (www.collegeboard.com).
 * copyright&copy; 2005 Cay S. Horstmann (http://horstmann.com)
 *
 * @author Alyce Brady
 * @author Jeff Raab, Northeastern University
 * @author Cay Horstmann
 */
public class DisplayMap
{
   // Class Variables: Not tied to any one DisplayMap object
   private HashMap<Class, Display> map = new HashMap<Class, Display>(); // the
   // collection
   private Display defaultDisplay = new DefaultDisplay();

   /**
    * Associates a display object with an environment object class. If no class
    * can be found for that name, an error message is printed.
    * 
    * @param envObjClassname
    *           the name of a class whose objects might be in an
    *           <code>Environment</code>
    * @param displayObj
    *           object that knows how to display objects of class
    *           <code>envObjClassname</code>
    */

   private Display createDisplay(Class cl)
   {
      try
      {
         ImageDisplay display = new ImageDisplay(cl);
         map.put(cl, display);
         return display;
      }
      catch (Exception e)
      {
         // oh well...
      }

      try
      {
         String className = cl.getName();
         Class dcl = Class.forName(className + "Display");
         if (Display.class.isAssignableFrom(cl))
         {
            Display display = (Display) dcl.newInstance();
            map.put(cl, display);
            return display;
         }
      }
      catch (Exception e)
      {
         // oh well...
      }
      return null;
   }

   /**
    * Finds a display class that knows how to display the given object.
    * 
    * @param obj
    *           the object to display
    */
   public Display findDisplayFor(Class cl)
   {
      // Go up through the class hierarchy for obj and see
      // if there is a display for its class or superclasses.

      if (cl == Object.class) return defaultDisplay;
      Display display = map.get(cl);
      if (display != null) return display;
      display = createDisplay(cl);
      if (display != null)
      {
         map.put(cl, display);
         return display;
      }
      display = findDisplayFor(cl.getSuperclass());
      map.put(cl, display);
      return display;
   }
   
   public Icon getIcon(Class cl, int w, int h)
   {
      return new DisplayIcon(cl, w, h);
   }

   private class DisplayIcon implements Icon
   {
      private Display displayObj;
      private int width, height;

      public DisplayIcon(Class cl, int w, int h)
      {
         displayObj = findDisplayFor(cl);
         width = w;
         height = h;
      }
      public int getIconWidth()
      {
         return width;
      }
      public int getIconHeight()
      {
         return height;
      }
      public void paintIcon(Component comp, Graphics g, int x, int y)
      {
         Graphics2D g2 = (Graphics2D) g;
         AffineTransform savedTransform = g2.getTransform(); // save current
         displayObj.draw(null, comp, g2, new Rectangle(x, y, getIconWidth(),
               getIconHeight()));
         g2.setTransform(savedTransform); // restore coordinate system
      }
   }
}
