package com.collegeboard.gridworld.gui;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.ResourceBundle;

import javax.swing.JPanel;
import javax.swing.JToolTip;
import javax.swing.JViewport;
import javax.swing.Scrollable;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.Timer;
import javax.swing.ToolTipManager;

import com.collegeboard.gridworld.grid.Grid;
import com.collegeboard.gridworld.grid.Location;

/**
 * AP&reg; Computer Science Marine Biology Simulation:<br>
 * A <code>MBSDisplay</code> is a panel containing a graphical display of the
 * environment of fish.
 * 
 *  <p>
 *  copyright&copy; 2002 College Entrance Examination Board (www.collegeboard.com).
 *  copyright&copy; 2005 Cay S. Horstmann (http://horstmann.com)
 *
 *  @author Julie Zelenski
 *  @author Cay Horstmann
 */

public class GridPanel extends JPanel
      implements Scrollable, PseudoInfiniteViewport.Pannable
{
   // Class constants
   private static final int MIN_CELL_SIZE = 8, DEFAULT_CELL_SIZE = 16;

   // Encapsulated data used to monitor/display the environment
   private Grid<?> grid;
   private int numRows, numCols, originRow, originCol;
   private int cellSize;
   private boolean toolTipsEnabled;
   private Color backgroundColor = Color.WHITE;
   private ResourceBundle resources;
   private DisplayMap displayMap;
   private Location currentLocation;
   private Timer tipTimer;
   private JToolTip tip;
   private JPanel glassPane;
   private static final int TIP_DELAY = 1000;

   /**
    * Construct a new MBSDisplay object with no environment. The view will be
    * empty.
    */
   public GridPanel(DisplayMap map, ResourceBundle res)
   {
      displayMap = map;
      resources = res;
   }
   
   /**
    * Paint this component.
    * 
    * @param g
    *           the Graphics object to use to render this component
    */
   public void paintComponent(Graphics g)
   {
      Graphics2D g2 = (Graphics2D) g;

      super.paintComponent(g2);
      if (grid == null)
         return;

      Insets insets = getInsets();
      g2.setColor(backgroundColor); // fill blue background
      g2.fillRect(insets.left, insets.top, numCols * cellSize, numRows
            * cellSize);

      drawGridlines(g2);

      ArrayList<Location> theActorLocations = grid.getOccupiedLocations();
      for (int index = 0; index < theActorLocations.size(); index++)
      {
         Location loc = (Location) theActorLocations.get(index);
         drawOccupant(g2, loc, grid.get(loc));
      }

   }

   /**
    * Draw one environment object. First verify that the object is actually
    * visible before any drawing, set up the clip appropriately and use the
    * DisplayMap to determine which object to call upon to render this
    * particular Locatable. Note that we save and restore the graphics transform
    * to restore back to normalcy no matter what the renderer did to to the
    * coordinate system.
    * 
    * @param g2
    *           the Graphics2D object to use to render
    * @param obj
    *           the Locatable object to draw
    */
   private void drawOccupant(Graphics2D g2, Location loc, Object obj)
   {
      Rectangle cellToDraw = new Rectangle(colToXCoord(loc.col()) + 1,
            rowToYCoord(loc.row()) + 1, cellSize - 1, cellSize - 1);

      // Only draw if the object is visible within the current clipping region.
      if (cellToDraw.intersects(g2.getClip().getBounds()))
      {
         AffineTransform savedTransform = g2.getTransform(); // save

         // Get the drawing object to display this environment object.
         Display displayObj = displayMap.findDisplayFor(obj.getClass());
         displayObj.draw(obj, this, g2, cellToDraw);

         g2.setTransform(savedTransform); // restore coordinate system
      }
   }

   /**
    * Draw the gridlines for the environment. We only draw the portion of the
    * lines that intersect the current clipping bounds.
    * 
    * @param g2
    *           the Graphics2 object to use to render
    */
   private void drawGridlines(Graphics2D g2)
   {
      Rectangle curClip = g2.getClip().getBounds();
      int top = getInsets().top, left = getInsets().left;

      int miny = Math.max(0, (curClip.y - top) / cellSize) * cellSize + top;
      int minx = Math.max(0, (curClip.x - left) / cellSize) * cellSize + left;
      int maxy = Math.min(numRows,
            (curClip.y + curClip.height - top + cellSize - 1) / cellSize)
            * cellSize + top;
      int maxx = Math.min(numCols,
            (curClip.x + curClip.width - left + cellSize - 1) / cellSize)
            * cellSize + left;

      g2.setColor(Color.black);
      g2.setStroke(new BasicStroke());
      for (int y = miny; y <= maxy; y += cellSize)
         // draw horizontal lines
         g2.drawLine(minx, y, maxx, y);

      for (int x = minx; x <= maxx; x += cellSize)
         // draw vertical lines
         g2.drawLine(x, miny, x, maxy);
      if (currentLocation != null)
      {         
         Point p = pointForLocation(currentLocation);
         g2.drawRect(p.x - cellSize / 2 - 1, p.y - cellSize / 2 - 1, cellSize + 2, cellSize + 2);
      }
   }

   /**
    * Enables/disables showing of tooltip giving information about the
    * environment object beneath the mouse.
    * 
    * @param flag
    *           whether to enable/disable tool tips
    */
   public void setToolTipsEnabled(boolean flag)
   {
      if (flag)
         ToolTipManager.sharedInstance().registerComponent(this);
      else
         ToolTipManager.sharedInstance().unregisterComponent(this);
      toolTipsEnabled = flag;
   }

   /**
    * Sets the Environment being displayed. Reset the cellSize to be the largest
    * that fits the entire environment in the current visible area (use default
    * if environment is too large).
    * 
    * @param newEnv
    *           the Environment to display
    */
   public void setGrid(Grid<?> newEnv)
   {
      currentLocation = new Location(0, 0);
      JViewport vp = getEnclosingViewport(); // before changing, reset
                                             // scroll/pan position
      if (vp != null)
         vp.setViewPosition(new Point(0, 0));

      grid = newEnv;
      originRow = originCol = 0;

      if (grid.numRows() == -1 && grid.numCols() == -1)
      {
         numRows = numCols = 2000; // This determines the "virtual" size of the
                                    // pan world
      }
      else
      {
         numRows = grid.numRows();
         numCols = grid.numCols();
      }
      recalculateCellSize(DEFAULT_CELL_SIZE);
   }

   // private helpers to caculate extra width/height needs for borders/insets.
   private int extraWidth()
   {
      return getInsets().left + getInsets().right + 1;
   }

   private int extraHeight()
   {
      return getInsets().top + getInsets().left + 1;
   }

   /**
    * Returns the desired size of the display, for use by layout manager.
    * 
    * @return preferred size
    */
   public Dimension getPreferredSize()
   {
      return new Dimension(numCols * cellSize + extraWidth(), numRows
            * cellSize + extraHeight());
   }

   /**
    * Returns the minimum size of the display, for use by layout manager.
    * 
    * @return minimum size
    */
   public Dimension getMinimumSize()
   {
      return new Dimension(numCols * MIN_CELL_SIZE + extraWidth(), numRows
            * MIN_CELL_SIZE + extraHeight());
   }

   /**
    * Zooms in the display by doubling the current cell size.
    */
   public void zoomIn()
   {
      cellSize *= 2;
      revalidate();
   }

   /**
    * Zooms out the display by halving the current cell size.
    */
   public void zoomOut()
   {
      cellSize = Math.max(cellSize / 2, MIN_CELL_SIZE);
      revalidate();
   }
 
   /**
    * Pans the display back to the origin, so that 0, 0 is at the the upper left
    * of the visible viewport.
    */
   public void recenter(Location loc)
   {
      originRow = loc.row();
      originCol = loc.col();
      repaint();
      JViewport vp = getEnclosingViewport();
      if (vp != null)
      {
         if (!isPannableUnbounded() || !(vp instanceof PseudoInfiniteViewport))
            vp.setViewPosition(pointForLocation(loc));
         else
            showPanTip();
      }
   }

   /**
    * Given a Point determine which environment location (if any) is under the
    * mouse. This method is used by the GUI when creating Fish by clicking on
    * cells in the display.
    * 
    * @param p
    *           the Point in question (in display's coordinate system)
    * @return the environment Location beneath the event, or null if no cell is
    *         beneath the mouse
    */
   public Location locationForPoint(Point p)
   {
      Location loc = new Location(yCoordToRow(p.y), xCoordToCol(p.x));
      return (grid != null && grid.isValid(loc)) ? loc : null;
   }

   public Point pointForLocation(Location loc)
   {
      return new Point(colToXCoord(loc.col()) + cellSize / 2, rowToYCoord(loc.row()) + cellSize / 2);
   }
   
   // private helpers to convert between (x,y) and (row,col)
   private int xCoordToCol(int xCoord)
   {
      return (xCoord - getInsets().left) / cellSize + originCol;
   }
   private int yCoordToRow(int yCoord)
   {
      return (yCoord - getInsets().top) / cellSize + originRow;
   }
   private int colToXCoord(int col)
   {
      return (col - originCol) * cellSize + getInsets().left;
   }
   private int rowToYCoord(int row)
   {
      return (row - originRow) * cellSize + getInsets().top;
   }

   /**
    * Given a MouseEvent, determine what text to place in the floating tool tip
    * when the the mouse hovers over this location. If the mouse is over a valid
    * environment cell. we provide some information about the cell and its
    * contents. This method is automatically called on mouse-moved events since
    * we register for tool tips.
    * 
    * @param evt
    *           the MouseEvent in question
    * @return the tool tip string for this location
    */
   public String getToolTipText(MouseEvent evt)
   {
      Location loc = locationForPoint(evt.getPoint());
      if (!toolTipsEnabled || loc == null)
         return null;
      return getToolTipText(loc);
   }
      
   private String getToolTipText(Location loc)
   {
      Object f = grid.get(loc);
      if (f != null)
         return MessageFormat.format(resources
               .getString("cell.tooltip.nonempty"), new Object[]
            { loc, f });
      else
         return MessageFormat.format(resources
               .getString("cell.tooltip.empty"), new Object[]
            { loc, f });
   }

   public void setCurrentLocation(Location loc)
   {
      currentLocation = loc;
   }      
   
   public Location getCurrentLocation()
   {
      return currentLocation;
   }
   
   public void moveLocation(int dr, int dc)
   {
      Location newLocation = new Location(currentLocation.row() + dr, currentLocation.col() + dc);
      if (!grid.isValid(newLocation)) return;

      currentLocation = newLocation;

      JViewport viewPort = getEnclosingViewport();      
      if (isPannableUnbounded()) 
      {
         if (originRow > currentLocation.row()) originRow = currentLocation.row(); 
         if (originCol > currentLocation.col()) originCol = currentLocation.col(); 
         Dimension dim = viewPort.getSize();
         int rows = dim.height / cellSize;
         int cols = dim.width / cellSize;
         if (originRow + rows - 1 < currentLocation.row()) originRow = currentLocation.row() - rows + 1;
         if (originCol + rows - 1 < currentLocation.col()) originCol = currentLocation.col() - cols + 1;         
      }      
      else if (viewPort != null)
      {
         int dx = 0;
         int dy = 0;
         Point p = pointForLocation(currentLocation);
         Rectangle locRect = new Rectangle(p.x - cellSize / 2, p.y - cellSize / 2, cellSize, cellSize);
         
         Rectangle viewRect = viewPort.getViewRect();
         if (!viewRect.contains(locRect))
         {
            while (locRect.x < viewRect.x + dx) dx -= cellSize;
            while (locRect.y < viewRect.y + dy) dy -= cellSize;
            while (locRect.getMaxX() > viewRect.getMaxX() + dx) dx += cellSize;
            while (locRect.getMaxY() > viewRect.getMaxY() + dy) dy += cellSize;
            
            
            Point pt = viewPort.getViewPosition();
            pt.x += dx;
            pt.y += dy;
            viewPort.setViewPosition(pt);
         }
      }
      repaint();
      showTip(currentLocation.toString(), pointForLocation(currentLocation));
   }     

   public void showTip(String tipText, Point pt)
   {
       if (getRootPane() == null) return;
       // draw in glass pane to appear on top of other components
       if (glassPane == null) 
       {
           getRootPane().setGlassPane(glassPane = new JPanel());
           glassPane.setOpaque(false);
           glassPane.setLayout(null);  // will control layout manually
           glassPane.add(tip = new JToolTip());
           tipTimer = new Timer(TIP_DELAY, new ActionListener() {
               public void actionPerformed(ActionEvent evt) { glassPane.setVisible(false); }});
           tipTimer.setRepeats(false);
       }
       if (tipText == null) return;
       
       // set tip text to identify current origin of pannable view
       tip.setTipText(tipText);
       
       // position tip to appear at upper left corner of viewport
       tip.setLocation(SwingUtilities.convertPoint(this, pt, glassPane));
       tip.setSize(tip.getPreferredSize());
       
       // show glass pane (it contains tip)
       glassPane.setVisible(true);
       glassPane.repaint();
       
       // this timer will hide the glass pane after a short delay
       tipTimer.restart();
   }  
  
   /**
    * Calculate the cell size to use given the current viewable region and the
    * the number of rows and columns in the environment. We use the largest
    * cellSize that will fit in the viewable region, bounded to be at least the
    * parameter minSize.
    */
   private void recalculateCellSize(int minSize)
   {
      if (numRows == 0 || numCols == 0)
      {
         cellSize = 0;
      }
      else
      {
         JViewport vp = getEnclosingViewport();
         Dimension viewableSize = (vp != null) ? vp.getSize() : getSize();
         cellSize = Math.min((viewableSize.height - extraHeight()) / numRows,
               (viewableSize.width - extraWidth()) / numCols);
         cellSize = Math.max(cellSize, minSize);
      }
      revalidate();
   }

   // helper to get our parent viewport, if we are in one.
   private JViewport getEnclosingViewport()
   {
      Component parent = getParent();
      return (parent instanceof JViewport) ? (JViewport) parent : null;
   }

   // MBSDisplay implements the Scrollable interface to get nicer behavior in a
   // JScrollPane
   // The 5 methods below are the methods in that interface

   public int getScrollableUnitIncrement(Rectangle visibleRect,
         int orientation, int direction)
   {
      return cellSize;
   }

   public int getScrollableBlockIncrement(Rectangle visibleRect,
         int orientation, int direction)
   {
      if (orientation == SwingConstants.VERTICAL)
         return (int) (visibleRect.height * .9);
      else
         return (int) (visibleRect.width * .9);
   }

   public boolean getScrollableTracksViewportWidth()
   {
      return false;
   }

   public boolean getScrollableTracksViewportHeight()
   {
      return false;
   }

   public Dimension getPreferredScrollableViewportSize()
   {
      return new Dimension(420, 420);
   }

   // MBSDisplay implements the PseudoInfiniteViewport.Pannable interface to
   // play nicely with the pan behavior for unbounded view.
   // The 3 methods below are the methods in that interface.

   public void panBy(int hDelta, int vDelta)
   {
      originCol += hDelta / cellSize;
      originRow += vDelta / cellSize;
      repaint();
   }

   public boolean isPannableUnbounded()
   {
      return (grid != null && grid.numRows() == -1);
   }
   
    /** Shows a tool tip over the upper left corner of the viewport 
    *  with the contents of the pannable view's pannable tip text
    *  (typically a string identifiying the corner point). Tip is
    *  removed after a short delay.
    **/
   public void showPanTip()
   {
      String tipText = null;
      Point upperLeft = new Point(0, 0);
      JViewport vp = getEnclosingViewport();
      if (!isPannableUnbounded() && vp != null)
         upperLeft = vp.getViewPosition();
      Location loc = locationForPoint(upperLeft);
      if (loc != null) tipText = loc.toString();
      
      showTip(tipText, getLocation());
   }
}
