package com.collegeboard.gridworld.gui;

import java.awt.Color;
import java.awt.Component;
import java.awt.MediaTracker;
import java.awt.Image;
import java.awt.image.FilteredImageSource;
import java.awt.image.RGBImageFilter;
import java.awt.Graphics2D;
import java.awt.geom.Ellipse2D;
import java.io.File;
import java.io.FileNotFoundException;
import java.net.URL;

import javax.swing.ImageIcon;
import java.util.HashMap;

/**
 *  <p>
 *  copyright&copy; 2002 College Entrance Examination Board (www.collegeboard.com).
 *  copyright&copy; 2005 Cay S. Horstmann (http://horstmann.com)
 *
 *  @author Julie Zelenski
 *  @author Cay Horstmann
 */

public class ImageDisplay extends AbstractDisplay
{
   private ImageIcon icon;
   private Image originalImage;
   private HashMap<Color, Image> tintedVersions = new HashMap<Color, Image>();

   /**
    * Constructs an object that knows how to display a fish as an image. Looks
    * for the named file first in the jar file, then in the current directory.
    * If the named file is not found or the file is malformed, a colored circle
    * will be substituted instead.
    * 
    * @param imageFilename
    *           name of file containing image
    */
   public ImageDisplay(Class cl)
         throws FileNotFoundException
   {
      String imageFilename = cl.getName().replace('.', '/') + ".gif";            
      URL url = cl.getClassLoader().getResource(imageFilename);
      
      if (url != null)
         icon = new ImageIcon(url);
      else
      {
         String path = System.getProperty("user.dir") + File.separator
               + cl.getName().replace('.', File.separatorChar);
         File file = new File(path);
         if (file.exists())
            icon = new ImageIcon(path);
         else
            throw new FileNotFoundException();
      }
      originalImage = icon.getImage();
   }

   /**
    * Draws a unit-length fish facing North. This implementation draws a Fish
    * object by tinting, scaling, and rotating the image provided to the
    * constructor.
    * 
    * @param obj
    *           fish we want to draw
    * @param comp
    *           the component we're drawing on
    * @param g2
    *           drawing surface
    */
   public void draw(Object obj, Component comp, Graphics2D g2)
   {
      Color color;
      if (obj == null)
         color = Color.BLACK;
      else
         color = (Color) getProperty(obj, "color");
      if (icon.getImageLoadStatus() != MediaTracker.COMPLETE)
      {
         // Image failed to load, so fall back to default display.
         if (color != null)
            g2.setPaint(color);
         g2.fill(new Ellipse2D.Double(-.4, -.4, .8, .8));
         return;
      }

      // Scale to shrink or enlarge the image to fit the size 1x1 cell.
      g2.scale(1.0 / icon.getIconWidth(), 1.0 / icon.getIconHeight());

      // Compose image with color using an image filter.
      Image tinted = (Image) tintedVersions.get(color);
      if (tinted == null) // not cached, need new filter for color
      {
         FilteredImageSource src = new FilteredImageSource(originalImage
               .getSource(), new TintFilter(color));
         tinted = comp.createImage(src);
         // Cache tinted image in map by color, we're likely to need it again.
         tintedVersions.put(color, tinted);
      }

      icon.setImage(tinted);
      icon.paintIcon(comp, g2, -icon.getIconWidth() / 2,
            -icon.getIconHeight() / 2);
   }

   /**
    * An image filter class that tints colors based on the tint provided to the
    * constructor (the color of a fish).
    */
   private static class TintFilter extends RGBImageFilter
   {
      private int tintR, tintG, tintB;

      /** Constructs an image filter for tinting colors in an image. * */
      public TintFilter(Color color)
      {
         canFilterIndexColorModel = true;
         int rgb = color.getRGB();
         tintR = (rgb >> 16) & 0xff;
         tintG = (rgb >> 8) & 0xff;
         tintB = rgb & 0xff;
      }

      public int filterRGB(int x, int y, int argb)
      {
         // Separate pixel into its RGB coomponents.
         int alpha = (argb >> 24) & 0xff;
         int red = (argb >> 16) & 0xff;
         int green = (argb >> 8) & 0xff;
         int blue = argb & 0xff;

         // Use NTSC/PAL algorithm to convert RGB to grayscale.
         int lum = (int) (0.2989 * red + 0.5866 * green + 0.1144 * blue);

         // Interpolate along spectrum black->white with tint at midpoint
         double scale = Math.abs((lum - 128) / 128.0); // absolute distance
                                                         // from midpt
         int edge = lum < 128 ? 0 : 255; // going towards white or black?
         red = tintR + (int) ((edge - tintR) * scale); // scale from midpt to
                                                         // edge
         green = tintG + (int) ((edge - tintG) * scale);
         blue = tintB + (int) ((edge - tintB) * scale);
         return (alpha << 24) | (red << 16) | (green << 8) | blue;
      }

   }

}
