package com.collegeboard.gridworld.gui;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.text.MessageFormat;
import java.util.Comparator;
import java.util.ResourceBundle;
import java.util.Set;
import java.util.TreeSet;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.*;

import com.collegeboard.gridworld.grid.*;
import com.collegeboard.gridworld.world.World;

/**
 * Copyright&copy; 2002 College Entrance Examination Board (www.collegeboard.com).
 * Copyright&copy; 2005 Cay S. Horstmann (http://horstmann.com)
 * 
 * @author Julie Zelenski
 * @author Cay Horstmann
 */

public class SimController
{
   public static final int INDEFINITE = 0, FIXED_STEPS = 1, PROMPT_STEPS = 2;

   private static final int MIN_DELAY_MSECS = 10, MAX_DELAY_MSECS = 1000;
   private static final int INITIAL_DELAY = MIN_DELAY_MSECS
         + (MAX_DELAY_MSECS - MIN_DELAY_MSECS) / 2;

   private Timer timer;
   private JButton stepButton, runButton, stopButton;
   private JComponent controlPanel;
   private GridPanel display;
   private WorldFrame parentFrame;
   private int runOption, numStepsToRun, numStepsSoFar;
   private ResourceBundle resources;
   private DisplayMap displayMap;
   private boolean running;
   private Set<Class> occupantClasses;

   /**
    * Creates a new SimulationController tied to the specified display and gui
    * frame.
    * 
    * @param parent
    *           the MBSGUIFrame for the simulation window
    * @param disp
    *           the EnvDisplay to use for the simulation
    */
   public SimController(WorldFrame parent, GridPanel disp,
         DisplayMap displayMap, ResourceBundle res, ClassLoader loader)
   {
      resources = res;
      display = disp;
      parentFrame = parent;
      this.displayMap = displayMap;
      makeControls();

      occupantClasses = new TreeSet<Class>(new 
            Comparator<Class>()
            {
               public int compare(Class a, Class b)
               {
                  return a.getName().compareTo(b.getName());
               }
            });
      
      World<?> world = parentFrame.getWorld();
      Grid<?> gr = world.getGrid();
      for (Location loc : gr.getOccupiedLocations())
         occupantClasses.add(gr.get(loc).getClass());
      for (String name : world.getOccupantClasses()) 
         try
         {
            occupantClasses.add(loader.loadClass(name));
         }
         catch (Exception ex)
         {
            ex.printStackTrace();
         }      
      
      timer = new Timer(INITIAL_DELAY, new ActionListener()
      {
         public void actionPerformed(ActionEvent evt)
         {
            step();
         }
      });
      
      display.addMouseListener(new MouseAdapter()
            {
               public void mousePressed(MouseEvent evt)
               {
                  Location loc = display.locationForPoint(evt.getPoint());
                  if (loc != null && !isRunning())
                  {
                     display.setCurrentLocation(loc);
                     locationClicked();
                  }
               }
            });
      stop();
   }

   /**
    * Advances the simulation one step.
    */
   public void step()
   {
      parentFrame.getWorld().step();
      parentFrame.repaint();
      if (++numStepsSoFar == numStepsToRun)
         stop();
      Grid<?> gr = parentFrame.getWorld().getGrid();
      for (Location loc : gr.getOccupiedLocations())
         occupantClasses.add(gr.get(loc).getClass());
   }

   /**
    * Starts a timer to repeatedly step the simulation at the speed currently
    * indicated by the speed slider up Depending on the run option, it will
    * either step the the simulation some fixed number of steps or indefinitely
    * until stopped.
    */
   public void run()
   {
      if (runOption == PROMPT_STEPS)
         if (!promptUserForStepCount())
            return;
      display.setToolTipsEnabled(false); // hide tool tips while running
      parentFrame.setRunMenuItemsEnabled(false);
      stopButton.setEnabled(true);
      stepButton.setEnabled(false);
      runButton.setEnabled(false);
      numStepsSoFar = 0;
      timer.start();
      running = true;
   }

   /**
    * Stops any existing timer currently stepping the simulation.
    */
   public void stop()
   {
      display.setToolTipsEnabled(true);
      parentFrame.setRunMenuItemsEnabled(true);
      timer.stop();
      stopButton.setEnabled(false);
      runButton.setEnabled(true);
      stepButton.setEnabled(true);
      running = false;
   }
   
   public boolean isRunning()
   {
      return running;
   }

   /**
    * Changes the current run option to one of INDEFINITE, FIXED_STEPS, or
    * PROMPT_STEPS to specify the run behavior. The text of the Run button is
    * updated to show the current option. The run option in effect is returned.
    * This usually will be the same as newOption but in cases where the change
    * failed to go through (user cancelled dialog), the retun value will be the
    * existing unchanged option.
    * 
    * @param newOption
    *           the run option to use for the simuiation
    * @returns the run option settled on by the simulation.
    */
   public int setRunOption(int newOption)
   {
      switch (newOption)
      {
         case INDEFINITE :
            runButton.setText(resources.getString("button.run.indef"));
            numStepsToRun = INDEFINITE;
            break;
         case FIXED_STEPS :
            if (!promptUserForStepCount())
               return runOption; // leave option unchanged
            runButton.setText(MessageFormat.format(resources
                  .getString("button.run.fixed"), new Object[]
               { new Integer(numStepsToRun) }));
            break;
         case PROMPT_STEPS :
            runButton.setText(resources.getString("button.run.prompt"));
      }
      runOption = newOption;
      return runOption;
   }

   /**
    * Runs a dialog asking the user to input a number of steps for the run
    * behavior. If successful, updates numStepsToRun and returns true, else
    * returns false.
    */
   private boolean promptUserForStepCount()
   {
      int suggested = numStepsToRun > 0 ? numStepsToRun : 10;
      String response = (String) JOptionPane.showInputDialog(parentFrame,
            resources.getString("dialog.step.prompt"), resources
                  .getString("dialog.step.title"),
            JOptionPane.QUESTION_MESSAGE, null, null, "" + suggested);
      if (response != null)
         try
         {
            int userEntered = Integer.parseInt(response.trim());
            if (userEntered <= 0)
               throw new NumberFormatException();
            numStepsToRun = userEntered;
            return true;
         }
         catch (Exception ex)
         {
            Toolkit.getDefaultToolkit().beep();
         }
      return false;
   }

   /**
    * Builds the panel with the various simulation controls (buttons and
    * slider).
    */
   private void makeControls()
   {
      controlPanel = new JPanel();
      controlPanel.setLayout(new BoxLayout(controlPanel, BoxLayout.X_AXIS));
      controlPanel.setBorder(BorderFactory.createTitledBorder(resources
            .getString("button.sim.border")));
      controlPanel.add(Box.createRigidArea(new Dimension(5, 0)));

      controlPanel.add(stepButton = new JButton(resources
            .getString("button.sim.step")));
      controlPanel.add(runButton = new JButton(resources
            .getString("button.sim.run")));
      controlPanel.add(stopButton = new JButton(resources
            .getString("button.sim.stop")));
      runButton.setEnabled(false);
      stepButton.setEnabled(false);
      stopButton.setEnabled(false);

      controlPanel.add(Box.createRigidArea(new Dimension(5, 0)));
      controlPanel.add(new JLabel(resources.getString("slider.sim.slow")));
      JSlider speedSlider = new JSlider(MIN_DELAY_MSECS, MAX_DELAY_MSECS,
            INITIAL_DELAY);
      speedSlider.setInverted(true);
      speedSlider.setPreferredSize(new Dimension(100, speedSlider
            .getPreferredSize().height));
      speedSlider.setMaximumSize(speedSlider.getPreferredSize());

      controlPanel.add(speedSlider);
      controlPanel.add(new JLabel(resources.getString("slider.sim.fast")));
      controlPanel.add(Box.createRigidArea(new Dimension(5, 0)));

      stepButton.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent e)
         {
            step();
         }
      });
      runButton.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent e)
         {
            run();
         }
      });
      stopButton.addActionListener(new ActionListener()
      {
         public void actionPerformed(ActionEvent e)
         {
            stop();
         }
      });
      speedSlider.addChangeListener(new ChangeListener()
      {
         public void stateChanged(ChangeEvent evt)
         {
            timer.setDelay(((JSlider) evt.getSource()).getValue());
         }
      });
   }

   /**
    * Returns the panel containing the simulation controls.
    * 
    * @return the control panel
    */
   public JComponent controlPanel()
   {
      return controlPanel;
   }
   
   /**
    * Callback on mousePressed when editing an environment.
    */
   private void locationClicked()
   {
      World<?> world = parentFrame.getWorld();
      Location loc = display.getCurrentLocation();
      if (loc != null && !world.locationClicked(loc))
         editLocation();
      parentFrame.repaint();
   }   

   /**
    * Edits the contents of the current location, by displaying the
    * constructor or method menu.
    */
   public void editLocation()
   {
      World<?> world = parentFrame.getWorld();

      Location loc = display.getCurrentLocation();
      if (loc != null)
      {
         Object occupant = world.getGrid().get(loc);
         if (occupant == null)
         {
            MenuMaker maker = new MenuMaker(parentFrame, resources, displayMap);
            JPopupMenu popup = maker.makeOcupantConstructorMenu(occupantClasses, loc);
            Point p = display.pointForLocation(loc);
            popup.show(display, p.x, p.y);                                    
         }
         else
         {
            MenuMaker maker = new MenuMaker(parentFrame, resources, displayMap);
            JPopupMenu popup = maker.makeMethodMenu(occupant, loc);
            Point p = display.pointForLocation(loc);
            popup.show(display, p.x, p.y);                       
         }         
      }
      parentFrame.repaint();
   }
   
   /**
    * Edits the contents of the current location, by displaying the
    * constructor or method menu.
    */
   public void deleteLocation()
   {
      World<?> world = parentFrame.getWorld();
      Location loc = display.getCurrentLocation();
      if (loc != null)
      {
         world.remove(loc);
         parentFrame.repaint();
      }
   }
}
