package com.collegeboard.gridworld.world;

import java.util.Set;
import java.util.TreeSet;

import javax.swing.JFrame;

import com.collegeboard.gridworld.grid.BoundedGrid;
import com.collegeboard.gridworld.grid.Grid;
import com.collegeboard.gridworld.grid.Location;
import com.collegeboard.gridworld.gui.WorldFrame;

/**
 * 
 * A world is the mediator between a grid and the GridWorld GUI. 
 * <p> 
 * copyright&copy; 2005 Cay S. Horstmann (http://horstmann.com)
 * @author Cay Horstmann
 */
public class World<T>
{
   private Grid<T> gr;
   private Set<String> occupantClassNames;
   private Set<String> gridClassNames;
   private String message;
   
   private static final int DEFAULT_ROWS = 10;
   private static final int DEFAULT_COLS = 10;

   public World()
   {
      this(new BoundedGrid<T>(DEFAULT_ROWS, DEFAULT_COLS));
   }

   public World(Grid<T> g)
   {
      gr = g;
      gridClassNames = new TreeSet<String>();
      occupantClassNames = new TreeSet<String>();
      addGridClass("com.collegeboard.gridworld.grid.BoundedGrid");
      addGridClass("com.collegeboard.gridworld.grid.UnboundedGrid");
   }

   /**
    * Constructs and shows a frame for this world.
    */
   public void show()
   {
      JFrame frame = new WorldFrame(this);
      frame.setVisible(true);
   }
   /**
    * Gets the grid managed by this world.
    * @return the grid
    */
   public Grid<T> getGrid() { return gr; }
   /**
    * Sets the grid managed by this world.
    * @param newGrid the new grid
    */
   public void setGrid(Grid<T> newGrid) { gr = newGrid; }

   /**
    * Sets the message to be displayed in the world frame above the grid.
    * @param newValue the new message
    */
   public void setMessage(String newMessage) { message = newMessage; }
   /**
    * Gets the message to be displayed in the world frame above the grid.
    * @return the message
    */
   public String getMessage() { return message; }      

   /**
    * This method is called when the user clicks on the step button, or
    * when run mode has been activated by clicking the run button.
    */
   public void step() {}
   /**
    * This method is called when the user clicks on a location in the
    * WorldFrame.
    * 
    * @param loc the grid location that the user selected
    * @return true if the world consumes the click, or false if the 
    * GUI should invoke the Location->Edit menu action
    */
   public boolean locationClicked(Location loc) { return false; }

   /**
    * Adds an occupant at a random location
    * @param occupant the occupant to add
    */
   public void add(T occupant)
   {      
      int TRIES = 1000;
      int rows = getGrid().numRows();
      if (rows < 0) rows = DEFAULT_ROWS; // unbounded grid
      int cols = getGrid().numCols();
      if (cols < 0) cols = DEFAULT_COLS; // unbounded grid
      
      for (int i = 1; i <= TRIES; i++)
      {
         Location loc = new Location((int)(Math.random() * rows),
               (int)(Math.random() * cols));
         if (getGrid().get(loc) == null)
         {
            add(loc, occupant);
            return;
         }
      }
   }
   /**
    * Adds an occupant at a given location. (This method is supplied
    * so that beginners can make a world without having to know about
    * Location objects.)
    * @param loc the location
    * @param occupant the occupant to add
    */
   public void add(Location loc, T occupant) { getGrid().put(loc, occupant); }

   /**
    * Removes an occupant from a given location.
    * @param loc the location
    * @return the removed occupant, or null if the location was empty
    */
   public T remove(Location loc) { return getGrid().remove(loc); }

   /**
    * Adds a class to be shown in the "Set grid" menu.
    * @param className the name of the grid class
    */
   public void addGridClass(String className) { gridClassNames.add(className); }
   
   /**
    * Adds a class to be shown when clicking on an empty cell
    * @param className the name of the occupant class
    */
   public void addOccupantClass(String className) { occupantClassNames.add(className); }
   
   /**
    * Gets a set of grid classes that should be used by
    * the world frame for this world.
    * @return the set of grid class names
    */
   public Set<String> getGridClasses() { return gridClassNames; }   

   /**
    * Gets a set of occupant classes that should be used by
    * the world frame for this world.
    * @return the set of occupant class names
    */
   public Set<String> getOccupantClasses() { return occupantClassNames; }
}
